from agentdojo.default_suites.v1_1.travel.task_suite import task_suite as travel_task_suite
from agentdojo.default_suites.v1_1.travel.user_tasks import TravelUserTask  # noqa: F401 - Register tasks

__all__ = ["travel_task_suite"]
